/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef MESH_H
#define MESH_H

#include "BoundingBox.hh"
#include "Color.hh"
#include "Intersection.hh"
#include "Math.hh"
#include "MeshData.hh"
#include "Progress.hh"
#include "Sculpt.hh"
#include <list>
#include <map>
#include <set>
#include <string>
#include <vector>

namespace SharpConstruct
{
	namespace Brush
	{
		class EditData;
	}

	class RestoreOrderData
	{
	public:
		RestoreOrderData()
		: Enabled( false )
		{}

		void Reset()
		{
			Vertices.clear();
			Triangles.clear();
			Quads.clear();
		}

		bool Enabled;
		std::map< unsigned, unsigned > Vertices, Triangles, Quads;
	};
	class VisibleElementData
	{
	public:
		VisibleElementData()
		: Enabled( false ),
		  Vertices( 0 ),
		  Triangles( 0 ),
		  Quads( 0 )
		{}

		bool Enabled;
		unsigned Vertices;
		unsigned Triangles;
		unsigned Quads;

		RestoreOrderData RestoreOrder;
	};
	
	class StoredNormalData
	{
	public:
		StoredNormalData()
		: Set( false )
		{}

		Normal3D Normals[ 8 ];
		bool Set;
	};

/** Stores and manipulates vertex and polygon data.
 *
 * Provides storage for vertex and polygon data, as well as data used to
 * display the mesh, and data relating the vertex and polygon data. Active data
 * is kept involving the area of the mesh currently being edited, and the
 * edit configuration. Functions are provided for file input/output, interactive
 * editing, deformations, and retrieval of mesh information.
**/

class Mesh
{
public:
	/*Mesh( Point3DVector& locs,
	      Point3DVector& norms,
	      std::vector< Color >& cols,
	      Point3DVector& tnorms,
	      Point3DVector& qnorms,
	      std::vector< Triangle >& tris,
	      std::vector< Quad >& quads,
	      std::vector< PolygonUsers >& users,
	      EdgeData& edges,
	      bool& saved,
	      VisibleElementData& visel,
	      BoundingBox& bounds,
	      std::string& filename );*/
	
	Mesh( VertexData&,
	      ColorData&,
	      PolygonData&,
	      bool& saved,
	      VisibleElementData& visel,
	      BoundingBox& bounds,
	      std::string& filename );

	~Mesh()
	{}

	template< typename T > void Create( T creator, Progress& p )
	{
		if( T::RequiresInitialization )
			begin_creation_();
		finalize_creation_( creator( *this ), p );
	}

	// Basic file input/output control
	void Save( std::string filename );

	// Deformations
	void Move( float, float, float );
	void Resize( float, float, float );
	void Deflate( float v );
	void Spherize( float v );
	void Smooth( float );
	void Flood( const Color&, const float );
	void Subdivide( bool smooth );
	void AutomaticSubdivide();
	void AdaptiveSubdivide();
	void Optimize( float );
	int RemoveDuplicates();
	void Mirror( bool, bool, bool );
	void FlipNormals();
	void Unify();
	void ToTriangles();

	void HideBox( Point3DVector& hb );

	static void ClearModifiedVertexIndices();
	void RecalculateDamaged();
	void RecalculateAllNormals();
	void RecalculateModifiedNormals();

	Point3D AreaNormal( const Brush::EditData&, const std::vector< ActiveData >& ) const;
	void FindActiveVertices( std::vector< ActiveData >&,
				 const Brush::EditData& e );
	Point3D VertexNeighborLocAverage( unsigned v ) const;
	Color VertexNeighborClrAverage( unsigned v ) const;
	Intersections Intersects( const Point3D,
				  const Normal3D ) const;

	// Mesh information access
	void Clear();
	const Point3DVector& VertexLocations() const;
	Point3DVector& VertexLocations();
	const Point3DVector& VertexNormals() const;
	Point3DVector& VertexNormals();
	const std::vector< Color >& VertexColors() const;
	std::vector< Color >& VertexColors();

	float Mask( const unsigned ) const;

	const Point3DVector& TriangleNormals() const;
	Point3DVector& TriangleNormals();

	const Point3DVector& QuadNormals() const;
	Point3DVector& QuadNormals();

	const std::vector< Triangle >& Triangles() const;
	std::vector< Triangle >& Triangles();
	const std::vector< Quad >& Quads() const;
	std::vector< Quad >& Quads();
	const std::vector< PolygonUsers >& VertexUsers() const;
	std::vector< PolygonUsers >& VertexUsers();
	const EdgeData& Edges() const;
	EdgeData& Edges();
	const Grid& CurrentGrid() const;
	Grid& CurrentGrid();

	const Point3DVector& ProjectedLocations() const;
	Point3DVector& ProjectedLocations();

	static Point3D& CurrentSelection();
	static void SetSelectionRadius( const float );
	static float SelectionRadius();
	//typedef std::list< unsigned > DamagedGroup;
	typedef std::vector< unsigned > DamagedGroup;
	static DamagedGroup& DamagedTriangles();
	static DamagedGroup& DamagedQuads();
	static Normal3D& UpNormal();
	static Normal3D& RightNormal();
	static Normal3D& ZNormal();

	const VisibleElementData& VisibleElements() const;
	VisibleElementData& VisibleElements();

	typedef std::vector< unsigned > ModifiedGroup;
	static ModifiedGroup& ModifiedVertexIndices();

	void ChangeOperation( const bool, const bool );
	void MakeSane();
	  
	std::string FileName() const;
	void SetFileName( const std::string& );

	std::string GetModelStatisticsString() const;
	Brush::EditData GetMirror( bool x, bool y, bool z ) const;
	Math::PolarPoint Relative2DLocation( const Brush::EditData& e, const Point3D x, const float len ) const;
	void CopyEssentials( Mesh & );
	void ClearExtraData();
	
	void PerformanceTest( std::string test );

	void UpdateBounds();
	void UpdateBoundsPartial();
	const BoundingBox& Bounds() const;

	void ResetVisibles();
	void UpdateVertexUsers();

	Point3D PolygonCenter( const Triangle& );
	Point3D PolygonCenter( const Quad& );
private:
	void begin_creation_();
	void finalize_creation_( const bool, Progress& );

	void update_edges_();
	
	// File output
	void save_obj_( const std::string& );
	void save_wrl_( const std::string& );

	bool intersects_( const Triangle&,
			  const Point3D,
			  const Normal3D,
			  Point3D& ) const;
	bool intersects_( const Quad&,
			  const Point3D,
			  const Normal3D,
			  Point3D& ) const;

	class ActiveVertexData
	{
	public:
		ActiveVertexData( unsigned c, std::set< unsigned >& chk, std::vector< unsigned >& indices,
				  std::vector< float >& dists, Point3D& a )
			: Current( c ), Checked( chk ), VertexIndices( indices ),
			  VertexDistances( dists ), Area( a )
		{}

		unsigned Current;
		std::set< unsigned >& Checked;
		std::vector< unsigned >& VertexIndices;
		std::vector< float >& VertexDistances;
		Point3D& Area;
	};
	Point3D _get_mirror_point( bool, bool, bool );
	

	// Object data
	std::string& _filename;
	bool& _saved;
	
	unsigned _poly_count() const;
	void _clear_polys();
	
	VertexData& vertex_data_;
	ColorData& color_data_;
	PolygonData& polygon_data_;
	
	static DamagedGroup _damaged_triangles;
	static DamagedGroup _damaged_quads;
	
	std::map< unsigned, unsigned > _triangle_users( unsigned v1, unsigned v2 );
	std::map< unsigned, unsigned > _quad_users( unsigned v1, unsigned v2 );
	std::set< unsigned > _connected_polygon_vertices( unsigned v ) const;
	std::set< unsigned > _connected_vertices( unsigned v ) const;

	void add_edge_midpoint_( Edge& );

	void generate_subdivided_polygons_();
	
	// Active data
	static float _selection_radius;
	static Point3D _current_selection;
	static Normal3D _selections[ 8 ];
	static StoredNormalData area_normals_;
	static Normal3D _view_up_normal;
	static Normal3D _view_right_normal;
	static Normal3D _view_z_normal;
	static ModifiedGroup _modified_vertex_indices;
	static Point3DVector _selection_tube;
	static Point3DVector _projected_locations;
	
	VisibleElementData& _visible_elements;
	
	BoundingBox& bounds_;
};
}
#endif // MESH_H
