/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Brush.hh"
#include "MeshHistory.hh"
#include "Tools.hh"
#include <cmath>
#include <queue>

using namespace SharpConstruct;

Brush::EditData Mesh::GetMirror( bool x, bool y, bool z ) const
{
	Brush::EditData e;

	e.SymmetryIndex = x * 4 + y * 2 + z;
	e.Center = _current_selection;
	e.Up = _view_up_normal;
	e.Left = _view_right_normal;

	if( x )
	{
		e.Center.X() = -e.Center.X();
		e.Up.X() = -e.Up.X();
		e.Left.X() = -e.Left.X();
	}
	if( y )
	{
		e.Center.Y() = -e.Center.Y();
		e.Up.Y() = -e.Up.Y();
		e.Left.Y() = -e.Left.Y();
	}
	if( z )
	{
		e.Center.Z() = -e.Center.Z();
		e.Up.Z() = -e.Up.Z();
		e.Left.Z() = -e.Left.Z();
	}

	return e;
}

Math::PolarPoint Mesh::Relative2DLocation( const Brush::EditData& e, Point3D x, float len ) const
{
	const float PI = 3.1415;

	Point3D t2( x - e.Center );
	Normalize( t2 );

	float theta = HorizontalAdd( e.Left * t2 );

	// Avoid NaN errors
	if( theta < -1 )
		theta = -1;
	else if( theta > 1 )
		theta = 1;

	theta = acos( theta );

	/* Checks whether theta should be in the III/IV quadrants using the
	   dot product with the Up vector */
	if( HorizontalAdd( e.Up * t2 ) > 0 )
		theta = 2 * PI - theta;

	return Math::PolarPoint( len / _selection_radius, theta );
}

void Mesh::FindActiveVertices( std::vector< ActiveData >& active_data,
			       const Brush::EditData& e )
{
	active_data.clear();

	if( Distance( e.Center, _selections[ e.SymmetryIndex ] ) < _selection_radius )
		area_normals_.Set = true;
	else
	{
		area_normals_.Set = false;
		_selections[ e.SymmetryIndex ] = e.Center;
	}

	CurrentGrid().FindActiveCells( e );
	Grid::ActiveCellType& cells( CurrentGrid().ActiveCells() );

	// Turns the active area green for testing
	/*for( Grid::ActiveCellType::iterator i = cells.begin();
	     i != cells.end(); ++i )
	{
		const std::vector< unsigned >& indices( **i );
		for( unsigned i = 0; i < indices.size(); ++i )
			VertexColors()[ indices[ i ] ] = Color( 0, 1, 0 );
	}*/

	for( Grid::ActiveCellType::iterator i = cells.begin();
	     i != cells.end(); ++i )
	{
		const std::vector< unsigned >& indices( **i );
		const unsigned size( indices.size() );
		//std::cout << size << std::endl;
		for( unsigned i = 0; i < size; ++i )
		{
			const unsigned current( indices[ i ] );
			const float distance( Distance( e.Center, VertexLocations()[ current ] ) );
			if( distance < _selection_radius )
			{
				const float curve_val( Tools::Instance().BrushCurveEnabled() ?
						       Tools::Instance().BrushCurve().Data().Value( 1 - distance / _selection_radius ) :
						       1 );
				const float strength( Tools::Instance().Brush()->AlphaStrength( Relative2DLocation( e, VertexLocations()[ current ],
														    distance ) ) * curve_val );
				active_data.push_back( ActiveData( current, strength ) );
			}
		}
	}

	for( std::vector< ActiveData >::iterator i = active_data.begin();
	     i != active_data.end(); ++i )
		_modified_vertex_indices.push_back( i->Index() );
}

/*#ifdef _MSC_VER
_declspec(noinline)
#endif
unsigned FindInitPoint( const VisibleElementData& _visible_elements, Point3DVector& _vertex_locations,
			EditData& e, const float _selection_radius )
#ifndef _MSC_VER
                        __attribute__ ((noinline))
#endif
                        ;

// I seperated this out for better profiling data, it should be made member of Mesh
unsigned FindInitPoint( const VisibleElementData& _visible_elements, Point3DVector& _vertex_locations,
			EditData& e, const float _selection_radius )
{
	unsigned area_index( 0 );
	float tmp_distance;

	if( _selection_radius == 0 )
		return 0;

	//std::cout << _selection_radius << std::endl;
	
	
	// The point of this loop is just to get one vertex that is within the Radius
	unsigned i;
	for( i = 0; i < _visible_elements.Vertices; i++ )
	{
		e.Center.FastDistance( _vertex_locations[ i ], tmp_distance );
		
		if( tmp_distance < _selection_radius )
		{
			area_index = i;
			break;
		}
	}
	return area_index;
}

void Mesh::FindActiveVertices( std::vector< ActiveData >& active_data,
			       Brush::EditData& e )
{
	active_data.clear();

	unsigned area_index = 0;

	if( _selected_vertex[ e.SIndex ] >= _vertex_locations.size() )
		_selected_vertex[ e.SIndex ] = 0;

	// If a point used last time through the loop is within the Radius, use it!
	if( _vertex_locations[ _selected_vertex[ e.SIndex ] ].Distance( e.Center ) <
	    _selection_radius )
	{
		area_index = _selected_vertex[ e.SIndex ];
	}
	else
	{
		area_index = FindInitPoint( _visible_elements, _vertex_locations, e, _selection_radius );
	}

	// Using a vector seems to be at least slightly faster than a set
	std::vector< char > checked( _vertex_locations.size(), false );
	
	std::queue< unsigned > untested;

	// Load in the point we already know is inside...
	untested.push( area_index );

	while( !untested.empty() )
	{
		const unsigned current( untested.front() );
		untested.pop();

		if( checked[ current ] )
			continue;
		checked[ current ] = true;
		
		const float distance( e.Center.Distance( _vertex_locations[ current ] ) );
		if( distance < _selection_radius )
		{
			const float strength( CurrentBrush().AlphaStrength( Relative2DLocation( e, VertexLocations()[ current ],
												distance ) ) );
			active_data.push_back( ActiveData( current, strength ) );
		}
		else
			continue;
		
		// Loop through each triangle using the current point
		for( unsigned i = 0; i < _vertex_users[ current ].Triangles.size(); ++i )
		{
			// Set the current polygon
			const unsigned polygon_index( _vertex_users[ current ].Triangles[ i ] );
			if( polygon_index < _visible_elements.Triangles )
			{
				// Loop through each vertex in the polygon
				for( unsigned j = 0; j <= 2; ++j )
				{
					// If the current active point equals the current VertexIndex...
					if( _triangles[ polygon_index ].V[ j ] == current )
					{
						const unsigned add( j + 1 );
						const unsigned next_point_index( _triangles[ polygon_index ].V[ add > 2 ? 0 : add ] );
						if( next_point_index < _visible_elements.Vertices )
							untested.push( next_point_index );
						break;
					}
				}
			}
		}

		// Loop through each quad using the current point
		for( unsigned i = 0; i < _vertex_users[ current ].Quads.size(); ++i )
		{
			// Set the current polygon
			const unsigned polygon_index( _vertex_users[ current ].Quads[ i ] );
			if( polygon_index < _visible_elements.Quads )
			{
				// Loop through each vertex in the polygon
				for( unsigned j = 0; j <= 3; ++j )
				{
					// If the current active point equals the current VertexIndex...
					if( _quads[ polygon_index ].V[ j ] == current )
					{
						const unsigned add( j + 1 );
						const unsigned next_point_index( _quads[ polygon_index ].V[ add > 3 ? 0 : add ] );
						if( next_point_index < _visible_elements.Vertices )
							untested.push( next_point_index );
						break;
					}
				}
			}
		}
	}
	/////

	if( active_data.size() > 0 )
		_selected_vertex[ e.SIndex ] = active_data[ 0 ].Index();

	for( std::vector< ActiveData >::iterator i = active_data.begin();
	     i != active_data.end(); ++i )
		_modified_vertex_indices.push_back( i->Index() );

	//std::cout << active_data.size() << std::endl;
}*/

void Mesh::AdaptiveSubdivide()
{

	/*const unsigned tri_size( DamagedTriangles().size() );

	float largest( 0 );
	for( unsigned i = 0; i < tri_size; ++i )
	{
		Triangle& poly( Triangles()[ DamagedTriangles()[ i ] ] );
		const float area( 0.5 * Point3D::Origin().Distance( Cross( VertexLocations()[ poly.V[ 0 ] ],
										      VertexLocations()[ poly.V[ 1 ] ] ) *
								    Cross( VertexLocations()[ poly.V[ 0 ] ],
										      VertexLocations()[ poly.V[ 2 ] ] ) ) );
		if( area > largest )
			largest = area;
	}
	const float half_large( largest / 2.0f );

	for( unsigned i = 0; i < tri_size; ++i )
	{
		Triangle& poly( Triangles()[ DamagedTriangles()[ i ] ] );
		const float area( 0.5 * Point3D::Origin().Distance( Cross( VertexLocations()[ poly.V[ 0 ] ],
										      VertexLocations()[ poly.V[ 1 ] ] ) *
								    Cross( VertexLocations()[ poly.V[ 0 ] ],
										      VertexLocations()[ poly.V[ 2 ] ] ) ) );
		if( area > half_large )
		{
			const unsigned cntr_ndx( VertexLocations().size() );
			VertexLocations().push_back( _polygon_center( poly ) );
			DamagedTriangles().push_back( Triangles().size() );
			Triangles().push_back( Triangle( poly.V[ 0 ], poly.V[ 1 ], cntr_ndx ) );
			DamagedTriangles().push_back( Triangles().size() );
			Triangles().push_back( Triangle( poly.V[ 1 ], poly.V[ 2 ], cntr_ndx ) );
			poly = Triangle( poly.V[ 2 ], poly.V[ 0 ], cntr_ndx );
		}
	}*/

	/*VertexNormals().resize( VertexLocations().size() );
	VertexColors().resize( VertexLocations().size() );
	TriangleNormals().resize( Triangles().size() );
	QuadNormals().resize( Quads().size() );

	ResetVisibles();
	UpdateVertexUsers();

	RecalculateAllNormals();*/
}
