/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Brush.hh"
#include "Mesh.hh"
#include "MeshHistory.hh"
#include "Polygon3D.hh"
#include "Save.hh"
#include "Utilities.hh"
#include <cmath>
#include <fstream>
#include <queue>
#include <sstream>
#include <string>

#include <iomanip>
#include <iostream>

using namespace SharpConstruct;

Normal3D Mesh::_view_up_normal( 0, 0, 0 );
Normal3D Mesh::_view_right_normal( 0, 0, 0 );
Normal3D Mesh::_view_z_normal( 0, 0, 0 );
float Mesh::_selection_radius = 0;
Point3D Mesh::_current_selection( 0, 0, 0 );
Mesh::ModifiedGroup Mesh::_modified_vertex_indices;
Mesh::DamagedGroup Mesh::_damaged_triangles;
Mesh::DamagedGroup Mesh::_damaged_quads;
Point3D Mesh::_selections[ 8 ];
Point3DVector Mesh::_selection_tube;
Point3DVector Mesh::_projected_locations;
StoredNormalData Mesh::area_normals_;

/*Mesh::Mesh( Point3DVector& locs,
	    Point3DVector& norms,
	    std::vector< Color >& cols,
	    Point3DVector& tnorms,
	    Point3DVector& qnorms,
	    std::vector< Triangle >& tris,
	    std::vector< Quad >& quads,
	    std::vector< PolygonUsers >& users,
	    EdgeData& edges,
	    bool& saved,
	    VisibleElementData& visel,
	    BoundingBox& bb,
	    std::string& filename )
: _filename( filename ),
  _saved( saved ),
  VertexLocations()( locs ),
  VertexNormals()( norms ),
  VertexColors()( cols ),
  Triangles()( tris ),
  Quads()( quads ),
  TriangleNormals()( tnorms ),
  QuadNormals()( qnorms ),
  VertexUsers()( users ),
  Edges()( edges ),
  _visible_elements( visel ),
  bounds_( bb )
{}*/
Mesh::Mesh( VertexData& vd,
	    ColorData& cd,
	    PolygonData& pd,
	    bool& saved,
	    VisibleElementData& visel,
	    BoundingBox& bounds,
	    std::string& filename )
	: _filename( filename ),
	  _saved( saved ),
	  vertex_data_( vd ),
	  color_data_( cd ),
	  polygon_data_( pd ),
	  _visible_elements( visel ),
	  bounds_( bounds )
{}

/** Saves any supported mesh type.
 *
 * Parses the filename string to find the extension, then calls the appropriate
 * function to do the actual saving.
**/
void Mesh::Save( std::string filename )
{
	const std::string extension( ExtractExtension( filename ) );

	if( extension == "OBJ" )
		save_obj_( filename );
	else if( extension == "WRL" )
		save_wrl_( filename );
	else if( extension == "LWO" )
		SaveLWO( *this, filename );
	else
		save_obj_( filename ); // This should never happen.
	
	_filename = filename;
	_saved = true;
	_visible_elements.RestoreOrder.Enabled = true;
}

void Mesh::HideBox( Point3DVector& hb )
{
	Point3DVector vis, inv;
	std::vector< Color > colvis, colinv;
	std::map< unsigned, unsigned > mpvis, mpinv;
	std::vector< Triangle > vistris, invtris;
	std::vector< Quad > visquads, invquads;

	RestoreOrderData& ro = _visible_elements.RestoreOrder;

	Point3DVector n;
	n.resize( 4 );
	n[ 0 ] = PlaneNormal( hb[ 0 ], hb[ 1 ], hb[ 4 ] );
	n[ 1 ] = PlaneNormal( hb[ 1 ], hb[ 2 ], hb[ 5 ] );
	n[ 2 ] = PlaneNormal( hb[ 2 ], hb[ 3 ], hb[ 5 ] );
	n[ 3 ] = PlaneNormal( hb[ 3 ], hb[ 0 ], hb[ 4 ] );

	float d[ 4 ];
	for( unsigned i = 0; i <= 3; ++i )
		d[ i ] = hb[ i ].X() * n[ i ].X() + hb[ i ].Y() * n[ i ].Y() + hb[ i ].Z() * n[ i ].Z();

	for( unsigned i = 0; i < VertexLocations().size(); ++i )
	{
		// Test vertex's location with each of the four planes
		bool inside = true;
		for( unsigned j = 0; j <= 3; ++j )
		{
			if( VertexLocations()[ i ].X() * n[ j ].X() +
			    VertexLocations()[ i ].Y() * n[ j ].Y() +
			    VertexLocations()[ i ].Z() * n[ j ].Z() < d[ j ] )
			{
				inside = false;
				break;
			}
		}

		// If it's inside OR is already hidden, add to the hidden map
		if( inside || i >= _visible_elements.Vertices )
		{
			mpinv[ i ] = inv.size();
			inv.push_back( VertexLocations()[ i ] );
			colinv.push_back( VertexColors()[ i ] );
		}
		else // Add to the visible map
		{
			mpvis[ i ] = vis.size();
			vis.push_back( VertexLocations()[ i ] );
			colvis.push_back( VertexColors()[ i ] );
		}
	}

	_visible_elements.Vertices = vis.size();

	// Copy the vertices back, visible first, then hidden
	VertexLocations() = vis;
	for( unsigned i = 0; i < inv.size(); ++i )
		VertexLocations().push_back( inv[ i ] );

	// Same for colors
	VertexColors() = colvis;
	VertexColors().insert( VertexColors().end(), colinv.begin(), colinv.end() );

	for( unsigned i = 0; i < Triangles().size(); ++i )
	{
		bool visible = true;
		for( unsigned j = 0; j <= 2; ++j )
		{
			if( mpinv.find( Triangles()[ i ].V[ j ] ) != mpinv.end() )
			{
				visible = false;
				Triangles()[ i ].V[ j ] = mpinv[ Triangles()[ i ].V[ j ] ] + vis.size();
			}
			else
				Triangles()[ i ].V[ j ] = mpvis[ Triangles()[ i ].V[ j ] ];

		}
		if( visible )
			vistris.push_back( Triangles()[ i ] );
		else
			invtris.push_back( Triangles()[ i ] );
	}
	_visible_elements.Triangles = vistris.size();
	Triangles() = vistris;
	Triangles().insert( Triangles().end(), invtris.begin(), invtris.end() );

	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		bool visible = true;
		for( unsigned j = 0; j <= 3; ++j )
		{
			if( mpinv.find( Quads()[ i ].V[ j ] ) != mpinv.end() )
			{
				visible = false;
				Quads()[ i ].V[ j ] = mpinv[ Quads()[ i ].V[ j ] ] + vis.size();
			}
			else
				Quads()[ i ].V[ j ] = mpvis[ Quads()[ i ].V[ j ] ];

		}
		if( visible )
			visquads.push_back( Quads()[ i ] );
		else
			invquads.push_back( Quads()[ i ] );
	}
	_visible_elements.Quads = visquads.size();
	Quads() = visquads;
	Quads().insert( Quads().end(), invquads.begin(), invquads.end() );

	UpdateVertexUsers();
	RecalculateAllNormals();

	ro.Vertices = mpvis;
	for( std::map< unsigned, unsigned >::iterator i = mpinv.begin();
	     i != mpinv.end(); ++i )
		mpinv[ i->first ] = i->second + vis.size();

	ChangeOperation( true, true );
}

void Mesh::ChangeOperation( const bool upol, const bool all )
{
	_saved = false;
	MeshHistory::Instance().Changed()( upol, all );
}

void Mesh::MakeSane()
{
	VertexNormals().resize( VertexLocations().size() );
	VertexColors().resize( VertexLocations().size() );

	TriangleNormals().resize( Triangles().size() );
	QuadNormals().resize( Quads().size() );

	UpdateVertexUsers();

	RecalculateAllNormals();

	ResetVisibles();

	UpdateBounds();
	Mesh::CurrentGrid().SetDivision( 10 );
}

void Mesh::begin_creation_()
{
	_filename = "";
	Clear();
	// This prevents a crash that can happen in UpdateBoundsPartial() in some situations
        ClearModifiedVertexIndices();
}

void Mesh::finalize_creation_( const bool modifiable, Progress& p )
{
	VertexNormals().resize( VertexLocations().size() );
	VertexColors().resize( VertexLocations().size() );
	TriangleNormals().resize( Triangles().size() );
	QuadNormals().resize( Quads().size() );
	ResetVisibles();
	UpdateVertexUsers();
	if( modifiable )
	{
		Unify();
		Flood( Color( 1, 1, 1 ), 1 );
	}
	RecalculateAllNormals();
	_visible_elements.RestoreOrder.Reset();
	UpdateBounds();
	Mesh::CurrentGrid().SetDivision( 10 );
	ChangeOperation( true, true );
	_saved = true;
}

void Mesh::Clear()
{
	VertexLocations().clear();
	VertexNormals().clear();
	VertexColors().clear();
	_clear_polys();
}

const Point3DVector& Mesh::VertexLocations() const
{
	return vertex_data_.VertexLocations();
}
Point3DVector& Mesh::VertexLocations()
{
	return vertex_data_.VertexLocations();
}

const Point3DVector& Mesh::VertexNormals() const
{
	return vertex_data_.VertexNormals();
}
Point3DVector& Mesh::VertexNormals()
{
	return vertex_data_.VertexNormals();
}

const std::vector< Color >& Mesh::VertexColors() const
{
	return color_data_.VertexColors();
}
std::vector< Color >& Mesh::VertexColors()
{
	return color_data_.VertexColors();
}

float Mesh::Mask( const unsigned i ) const
{
	return 1.0f - VertexColors()[ i ].Mask();
}

const Point3DVector& Mesh::TriangleNormals() const
{
	return polygon_data_.TriangleNormals();
}
Point3DVector& Mesh::TriangleNormals()
{
	return polygon_data_.TriangleNormals();
}

const Point3DVector& Mesh::QuadNormals() const
{
	return polygon_data_.QuadNormals();
}
Point3DVector& Mesh::QuadNormals()
{
	return polygon_data_.QuadNormals();
}

const std::vector< Triangle >& Mesh::Triangles() const
{
	return polygon_data_.Triangles();
}
std::vector< Triangle >& Mesh::Triangles()
{
	return polygon_data_.Triangles();
}

const std::vector< Quad >& Mesh::Quads() const
{
	return polygon_data_.Quads();
}
std::vector< Quad >& Mesh::Quads()
{
	return polygon_data_.Quads();
}

const std::vector< PolygonUsers >& Mesh::VertexUsers() const
{
	return polygon_data_.VertexUsers();
}
std::vector< PolygonUsers >& Mesh::VertexUsers()
{
	return polygon_data_.VertexUsers();
}

const EdgeData& Mesh::Edges() const
{
	return polygon_data_.Edges();
}

EdgeData& Mesh::Edges()
{
	return polygon_data_.Edges();
}

const Grid& Mesh::CurrentGrid() const
{
	return polygon_data_.CurrentGrid();
}

Grid& Mesh::CurrentGrid()
{
	return polygon_data_.CurrentGrid();
}

const Point3DVector& Mesh::ProjectedLocations() const
{
	return _projected_locations;
}
Point3DVector& Mesh::ProjectedLocations()
{
	return _projected_locations;
}

Point3D& Mesh::CurrentSelection()
{
	return _current_selection;
}

void Mesh::SetSelectionRadius( const float r )
{
	_selection_radius = r;
}
float Mesh::SelectionRadius()
{
	return _selection_radius;
}

Mesh::DamagedGroup& Mesh::DamagedTriangles()
{
	return _damaged_triangles;
}
Mesh::DamagedGroup& Mesh::DamagedQuads()
{
	return _damaged_quads;
}

Normal3D& Mesh::UpNormal()
{
	return _view_up_normal;
}
Normal3D& Mesh::RightNormal()
{
	return _view_right_normal;
}
Normal3D& Mesh::ZNormal()
{
	return _view_z_normal;
}

const VisibleElementData& Mesh::VisibleElements() const
{
	return _visible_elements;
}
VisibleElementData& Mesh::VisibleElements()
{
	return _visible_elements;
}

const BoundingBox& Mesh::Bounds() const
{
	return bounds_;
}

Mesh::ModifiedGroup& Mesh::ModifiedVertexIndices()
{
	return _modified_vertex_indices;
}

std::string Mesh::FileName() const
{
	return _filename;
}
void Mesh::SetFileName( const std::string& f )
{
	_filename = f;
}

/** Returns mesh data information.
 *
 * Returns a string in this format:
 * "[number of vertices] vert[ex/ices], [number of polygons] polygon[s]"
**/
std::string Mesh::GetModelStatisticsString() const
{
	std::ostringstream oss;

	oss << VertexLocations().size();
	if( VertexLocations().size() == 1 )
		oss << " vertex, ";
	else
		oss << " vertices, ";
	oss << _poly_count();
	if( _poly_count() == 1 )
		oss << " polygon";
	else
		oss << " polygons";

	return oss.str();
}

void Mesh::CopyEssentials( Mesh &mesh )
{
	VertexLocations() = mesh.VertexLocations();
	VertexNormals() = mesh.VertexNormals();
	VertexColors() = mesh.VertexColors();
	TriangleNormals() = mesh.TriangleNormals();
	QuadNormals() = mesh.QuadNormals();
	Triangles() = mesh.Triangles();
	Quads() = mesh.Quads();
	VertexUsers() = mesh.VertexUsers();
	_filename = mesh._filename;
}

/** Free unimportant mesh data.
 *
 * Free memory by clearing temporary data unimportant to a mesh object.
**/
void Mesh::ClearExtraData()
{
	/*_active_vertex_indices.clear();
	  _active_vertex_distances.clear();*/
	//_modified_vertex_indices.clear();
}

void Mesh::PerformanceTest( std::string )
{
	/*CurrentBrush().StartSculpt( *this );
	for( unsigned i = 0; i < 25; ++i )
	{
		_current_selection = VertexLocations()[ rand() % VertexLocations().size() ];
		CurrentBrush().ContinueSculpt( *this );
	}
	//Unify();

	ChangeOperation( false, false );*/
}

/** Loads a mesh in the OBJ format.
 *
 * Loads mesh data from an OBJ file. Support is incomplete; the parsing has
 * bugs, and normal data is ignored.
**/
/*void Mesh::_load_obj( std::string filename )
{
	std::ifstream file( filename.c_str(), std::ios::in );
	//std::istringstream line;
	std::string token, subtoken;
	char buffer[ 256 ];
	unsigned polygon_count = 0;
	unsigned ndx;
	bool hascolor = false;

	VertexLocations().clear();
	VertexNormals().clear();
	VertexColors().clear();
	_uvs.clear();
	_clear_polys();
	std::vector< Polygon3D > polygons;

	while( !file.eof() )
	{
		std::istringstream line;
		file.getline( buffer, 256 );
		line.str( buffer );
		line >> token;
		if( token == "v" )
		{
			float x, y, z;
			line >> x >> y >> z;
			VertexLocations().push_back( Point3D( x, y, z ) );
		}
		else if( token == "vt" )
		{
			float u, v;
			line >> u >> v;
			_uvs.push_back( UV( u, v ) );
		}
		else if( token == "f" )
		{
			polygons.resize( polygon_count + 1 );
			while( line >> token )
			{
				int loc = token.find( '/' );
				subtoken = token.substr( 0, loc );
				polygons[ polygon_count ].VertexIndices.push_back( ToInt( subtoken ) - 1 );
			}
			polygon_count++;
		}
		else if( token == "#vertexcolor" )
		{
			line >> ndx;
			VertexColors().push_back( SharpConstruct::Color() );
			line >> VertexColors()[ ndx ].Red >> VertexColors()[ ndx ].Green >>
			        VertexColors()[ ndx ].Blue;
			hascolor = true;
		}
	}
	file.close();
	
	VertexNormals().resize( VertexLocations().size() );
	
	for( unsigned i = 0; i < polygons.size(); ++i )
	{
		if( polygons[ i ].VertexIndices.size() == 3 )
		{
			Triangles().push_back( Triangle( polygons[ i ].VertexIndices[ 0 ],
			                                polygons[ i ].VertexIndices[ 1 ],
			                                polygons[ i ].VertexIndices[ 2 ] ) );
		}
		else if( polygons[ i ].VertexIndices.size() == 4 )
		{
			Quads().push_back( Quad( polygons[ i ].VertexIndices[ 0 ],
			                        polygons[ i ].VertexIndices[ 1 ],
			                        polygons[ i ].VertexIndices[ 2 ],
			                        polygons[ i ].VertexIndices[ 3 ] ) );
		}
		else
		{
			for( unsigned j = 1; j < polygons[ i ].VertexIndices.size() - 1; ++j )
			{
				Triangles().push_back( Triangle( polygons[ i ].VertexIndices[ 0 ],
				                                polygons[ i ].VertexIndices[ j ],
				                                polygons[ i ].VertexIndices[ j + 1 ] ) );
			}
		}
	}
	
	TriangleNormals().resize( Triangles().size() );
	QuadNormals().resize( Quads().size() );
	
	UpdateVertexUsers();
	RecalculateAllNormals();
	
	if( !hascolor )
	{
		VertexColors().resize( VertexLocations().size() );
		Flood( 1 );
	}
	
	_saved = true;
}*/

/** Saves a mesh in the OBJ format.
 *
 * Saves mesh data to an OBJ file. Only vertex and polygon data is exported,
 * not normals. The top line is a comment including the SharpConstruct URL.
**/
void Mesh::save_obj_( const std::string& filename )
{
	std::ofstream file( filename.c_str(), std::ios::out );
	
	file << "# Exported from SharpConstruct " + SharpConstruct::GetVersion() +
	        " - http://sharp3d.sourceforge.net/\n\n";
	
	for( unsigned i = 0; i < VertexLocations().size(); ++i )
		file << "v " << VertexLocations()[ i ].X() << ' ' <<
	                    VertexLocations()[ i ].Y() << ' ' <<
	                    VertexLocations()[ i ].Z() << '\n';
	
	//for( i = 0; i < VertexLocations().size(); ++i )
	//	file << "vt " << _vertices.U[ i ] << ' ' << _vertices.V[ i ] << '\n';
	
	for( unsigned i = 0; i < Triangles().size(); ++i )
	{
		file << "f ";
		for( int j = 2; j >= 0; --j )
			file << Triangles()[ i ].V[ j ] + 1 << ' ';		
		file << '\n';
	}
	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		file << "f ";
		for( int j = 3; j >= 0; --j )
			file << Quads()[ i ].V[ j ] + 1 << ' ';		
		file << '\n';
	}
	
	for( unsigned i = 0; i < VertexColors().size(); ++i )
		file << "#vertexcolor " << i << ' ' << VertexColors()[ i ].R() << ' '
		     << VertexColors()[ i ].G() << ' ' << VertexColors()[ i ].B()
		     << '\n';
	
	file.close();
	
	_saved = true;
}

/** Saves a mesh in the VRML97 (.wrl) format.
 *
 * Saves mesh data to an WRL file. Only vertex and polygon data is exported,
 * not normals. The top line is the VRML magic numbers string,
 * followed by a line with a comment including the SharpConstruct URL.
 *
 * Thanks to Joerg Scheurich for this code.
**/
void Mesh::save_wrl_( const std::string& filename )
{
	std::ofstream file( filename.c_str(), std::ios::out );
	
	unsigned i, j;
	
	file << "#VRML V[ 1 ].0 utf8\n# Exported from SharpConstruct " + SharpConstruct::GetVersion() +
	        " - http://sharp3d.sourceforge.net/\n\n";
	file << "Transform\n";
	file << "  {\n";
	file << "  children\n";
	file << "    [\n";
	file << "    Shape\n";
	file << "      {\n";
	file << "      appearance       Appearance\n";
	file << "        {\n";
	file << "        material        Material\n";
	file << "          {\n";
	file << "          }\n";
	file << "        }\n";
	file << "      geometry       IndexedFaceSet\n";
	file << "        {\n";
	file << "        # double sided\n";
	file << "        solid FALSE\n";
	file << "        # single sided: inside or outside ?\n";
	file << "        #ccw TRUE\n";
	file << "        #ccw FALSE\n";
	file << "        # smooth if angle < 90 degree\n";
	file << "        creaseAngle 1.57\n";
	file << "        coord   Coordinate\n";
	file << "          {\n";
	file << "          point\n";
	file << "            [\n";
	for( i = 0; i < VertexLocations().size(); ++i )
	{
		file << "            " << 
		VertexLocations()[ i ].X() << ", " <<
		VertexLocations()[ i ].Y() << ", " <<
		VertexLocations()[ i ].Z() << ", \n";
	}
	file << "            ]\n";
	file << "          }\n";
	file << "        coordIndex\n";
	file << "          [\n";
	for( i = 0; i < Triangles().size(); ++i )
	{
		file << "          ";
		for( j = 0; j <= 2; ++j )
			file << Triangles()[ i ].V[ j ] << ", ";		
		file << " -1, " << '\n';
	}
	for( i = 0; i < Quads().size(); ++i )
	{
		file << "          ";
		for( j = 0; j <= 3; ++j )
			file << Quads()[ i ].V[ j ] << ", ";		
		file << " -1, " << '\n';
	}
	file << "          ]\n";

/* not implemented yet	
for( i = 0; i < _vertices.size(); ++i )
	file << "#vertexcolor " << i << ' ' << _vertices[ i ].R << ' ' <<
			_vertices[ i ].G << ' ' << _vertices[ i ].B << '\n';
*/
	file << "        }\n";
	file << "      }\n";
	file << "    ]\n";
	file << "  }\n";

	file.close();
	
	_saved = true;
}

Point3D Mesh::VertexNeighborLocAverage( unsigned v ) const
{
	Point3D average( 0, 0, 0 );
	unsigned total = 0;
	for( unsigned j = 0; j < VertexUsers()[ v ].Triangles.size(); ++j )
	{
		const unsigned polygon_index = VertexUsers()[ v ].Triangles[ j ];
		for( unsigned k = 0; k <= 2; k++ )
			if( Triangles()[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;
				if( next_point_index == 3 )
					next_point_index = 0;
				
				average += VertexLocations()[ v ];
				average += VertexLocations()[ Triangles()[ polygon_index ].V[ previous_point_index ] ];
				average += VertexLocations()[ Triangles()[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	for( unsigned j = 0; j < VertexUsers()[ v ].Quads.size(); ++j )
	{
		const unsigned polygon_index = VertexUsers()[ v ].Quads[ j ];
		for( unsigned k = 0; k <= 3; k++ )
			if( Quads()[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 3;
				if( next_point_index == 4 )
					next_point_index = 0;
				
				average += VertexLocations()[ v ];
				average += VertexLocations()[ Quads()[ polygon_index ].V[ previous_point_index ] ];
				average += VertexLocations()[ Quads()[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	
	average /= total;
	
	return average;
}
Color Mesh::VertexNeighborClrAverage( unsigned v ) const
{
	Color average( 0, 0, 0 );
	unsigned total = 0;
	for( unsigned j = 0; j < VertexUsers()[ v ].Triangles.size(); ++j )
	{
		const unsigned polygon_index = VertexUsers()[ v ].Triangles[ j ];
		for( unsigned k = 0; k <= 2; k++ )
			if( Triangles()[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;
				if( next_point_index == 3 )
					next_point_index = 0;
				
				average += VertexColors()[ v ];
				average += VertexColors()[ Triangles()[ polygon_index ].V[ previous_point_index ] ];
				average += VertexColors()[ Triangles()[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	for( unsigned j = 0; j < VertexUsers()[ v ].Quads.size(); ++j )
	{
		const unsigned polygon_index = VertexUsers()[ v ].Quads[ j ];
		for( unsigned k = 0; k <= 3; k++ )
			if( Quads()[ polygon_index ].V[ k ] == v )
			{
				int previous_point_index = k - 1;
				int next_point_index = k + 1;
				if( previous_point_index < 0 )
					previous_point_index = 3;
				if( next_point_index == 4 )
					next_point_index = 0;
				
				average += VertexColors()[ v ];
				average += VertexColors()[ Quads()[ polygon_index ].V[ previous_point_index ] ];
				average += VertexColors()[ Quads()[ polygon_index ].V[ next_point_index ] ];
				total += 3;
			}
	}
	
	average /= total;
	
	return average;
}

SharpConstruct::Intersections Mesh::Intersects( const Point3D l,
						const Normal3D d ) const
{
	Intersections hits;
	Point3D hit;
	for( unsigned i = 0; i < Triangles().size(); ++i )
	{
		const Triangle& poly( Triangles()[ i ] );
		if( intersects_( poly, l, d, hit ) )
		{
			hits.push_back( Intersection( hit, TriangleNormals()[ i ] ) );
		}
	}

	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		const Quad& poly( Quads()[ i ] );
		if( intersects_( poly, l, d, hit ) )
		{
			hits.push_back( Intersection( hit, QuadNormals()[ i ] ) );
		}
	}
	return hits;
}

// Uses code from http://www.lighthouse3d.com/opengl/maths/index.php?raytriint
bool Mesh::intersects_( const Triangle& poly,
			const Point3D orig,
			const Normal3D dir,
			Point3D& hit ) const
{
	const Point3D& v0( VertexLocations()[ poly.V[ 0 ] ] );
	const Point3D e1( VertexLocations()[ poly.V[ 1 ] ] - v0 );
	const Point3D e2( VertexLocations()[ poly.V[ 2 ] ] - v0 );

	const Point3D h( Cross( dir, e2 ) );
	const float a = Dot( e1, h );

	if (a > -0.00001 && a < 0.00001)
		return false;

	const float f = 1/a;
	const Point3D s( orig - v0 );
	const float u( f * (Dot(s,h)) );
	
	if (u < 0.0 || u > 1.0)
		return false;

	const Point3D q( Cross( s,e1 ) );
	const float v( f * Dot( dir, q ) );
	if (v < 0.0 || u + v > 1.0)
		return false;

	// at this stage we can compute t to find out where 
	// the intersection point is on the line
	const float t( f * Dot(e2,q) );
	if( t > 0 ) // ray intersection
	{
		hit = Point3D( VertexLocations()[ poly.V[ 0 ] ] * ( 1 - u - v ) +
			       VertexLocations()[ poly.V[ 1 ] ] * u +
			       VertexLocations()[ poly.V[ 2 ] ] * v );
		return true;
	}
	else // this means that there is a line intersection  
		 // but not a ray intersection
		 return false;
}

bool Mesh::intersects_( const Quad& poly,
			const Point3D orig,
			const Normal3D dir,
			Point3D& hit ) const
{
	if( intersects_( Triangle( poly.V[ 0 ], poly.V[ 1 ], poly.V[ 2 ] ),
			 orig, dir, hit ) )
		return true;
	
	return intersects_( Triangle( poly.V[ 2 ], poly.V[ 3 ], poly.V[ 0 ] ),
			    orig, dir, hit );
}

Point3D Mesh::PolygonCenter( const Triangle& t )
{
	Point3D c( 0, 0, 0 );
	for( unsigned i = 0; i <= 2; ++i )
		c += VertexLocations()[ t.V[ i ] ];
	c /= 3;
	return c;
}

Point3D Mesh::PolygonCenter( const Quad& q )
{
	Point3D c( 0, 0, 0 );
	for( unsigned i = 0; i <= 3; ++i )
		c += VertexLocations()[ q.V[ i ] ];
	c /= 4;
	return c;
}

/** Calculates center and diameter of sphere formed by vertices.
 *
 * All vertices in the model are used to create a virtual sphere, with a center
 * at the average of all the points and the diameter equal to two times the
 * distance from the center to the furthest point from the center.
**/
/*void Mesh::CalculateBoundingSphere()
{
	float tmp;

	_diameter = 0;
	
	_center.Zero();

	for( unsigned i = 0; i < VertexLocations().size(); ++i )
		_center += VertexLocations()[ i ];
	
	_center /= VertexLocations().size();
	
	for( unsigned i = 0; i < VertexLocations().size(); ++i )
	{
		tmp = _center.Distance( VertexLocations()[ i ] );
		if( tmp > _diameter )
			_diameter = tmp;
	}
	
	_diameter *= 2;
}*/

void Mesh::UpdateBounds()
{
	const unsigned size( VertexLocations().size() );
	if( !size )
		return;
	
	bounds_.Initialize( VertexLocations()[ 0 ] );
	for( unsigned i = 1; i < size; ++i )
		bounds_.Expand( VertexLocations()[ i ] );
}
void Mesh::UpdateBoundsPartial()
{
	for( unsigned i = 0; i < ModifiedVertexIndices().size(); ++i )
		bounds_.Expand( VertexLocations()[ ModifiedVertexIndices()[ i ] ] );
	CurrentGrid().UpdateCells();
}

/** Calculates the average normal of the area.
 *
 * Returns the average normal of all the points within the area.
**/
Point3D Mesh::AreaNormal( const Brush::EditData& e,
			  const std::vector< ActiveData >& active ) const
{
	if( !area_normals_.Set ||
	    area_normals_.Normals[ e.SymmetryIndex ].X() !=
	    area_normals_.Normals[ e.SymmetryIndex ].X() )
	{
		Point3DVector& norbackup = MeshHistory::Instance().GetPreviousComposite().Vertices()->VertexNormals();

		Normal3D area_normal;
		area_normal = Origin();
		
		for( unsigned i = 0; i < active.size(); ++i )
			area_normal += norbackup[ active[ i ].Index() ];
		
		Normalize( area_normal );

		area_normals_.Normals[ e.SymmetryIndex ] = area_normal;
		area_normals_.Set = true;
		
		return area_normal;
	}
	else
		return area_normals_.Normals[ e.SymmetryIndex ];

}

/** Updates all vertex and polygon normals.
 *
 * Acts like _recalculate_modified_normals(), but does the recalculation for
 * every vertex and every polygon.
 *
 * Note: all polygons are treated as planar
**/
void Mesh::RecalculateAllNormals()
{
	for( unsigned i = 0; i < TriangleNormals().size(); ++i )
	{
		TriangleNormals()[ i ] = PlaneUnormal
			( VertexLocations()[ Triangles()[ i ].V[ 0 ] ],
			  VertexLocations()[ Triangles()[ i ].V[ 1 ] ],
			  VertexLocations()[ Triangles()[ i ].V[ 2 ] ] );

	}
	for( unsigned i = 0; i < QuadNormals().size(); ++i )
	{
		QuadNormals()[ i ] = PlaneUnormal
			( VertexLocations()[ Quads()[ i ].V[ 0 ] ],
			  VertexLocations()[ Quads()[ i ].V[ 1 ] ],
			  VertexLocations()[ Quads()[ i ].V[ 2 ] ] );
	}

	unsigned j, k;
	for( unsigned i = 0; i < VertexUsers().size(); ++i )
	{
		Normal3D& cur_norm( VertexNormals()[ i ] );
		const PolygonUsers& usr( VertexUsers()[ i ] );
		cur_norm = Origin();
		for( j = 0; j < usr.Triangles.size(); ++j )
			cur_norm += TriangleNormals()[ usr.Triangles[ j ] ];
		for( k = 0; k < usr.Quads.size(); k++ )
			cur_norm += QuadNormals()[ usr.Quads[ k ] ];

		Normalize( cur_norm );
	}
}

void Mesh::ClearModifiedVertexIndices()
{
	//fill( _modified_vertex_indices.begin(), _modified_vertex_indices.end(), false );
	_modified_vertex_indices.clear();
}

void Mesh::RecalculateDamaged()
{
	/* Speed difference is undet. */

	/*_damaged_triangles.resize( VertexLocations().size() );
	_damaged_quads.resize( VertexLocations().size() );

	fill( _damaged_quads.begin(), _damaged_quads.end(), false );
	fill( _damaged_triangles.begin(), _damaged_triangles.end(), false );

	const unsigned mod_size( _modified_vertex_indices.size() );
	for( unsigned i = 0; i < mod_size; ++i )
	{
		const unsigned mod_ndx( _modified_vertex_indices[ i ] );

		for( unsigned j = 0; j < VertexUsers()[ mod_ndx ].Triangles.size(); ++j )
			_damaged_triangles[ VertexUsers()[ mod_ndx ].Triangles[ j ] ] = true;
		for( unsigned j = 0; j < VertexUsers()[ mod_ndx ].Quads.size(); ++j )
			_damaged_quads[ VertexUsers()[ mod_ndx ].Quads[ j ] ] = true;
	}

	return;*/

	// This actually works faster with vector than list,
	// and using list with sort+unique is even slower
	// This comment ST approved

	_damaged_triangles.clear();
	_damaged_quads.clear();

#ifdef BM
	for( unsigned i = 0; i < _modified_vertex_indices.size(); ++i )
	{
		if( _modified_vertex_indices[ i ] )
		{
			for( unsigned j = 0; j < VertexUsers()[ i ].Quads.size(); ++j )
				_damaged_quads.push_back( VertexUsers()[ i ].Quads[ j ] );
		}
	}
#else	
	for( unsigned i = 0; i < _modified_vertex_indices.size(); ++i )
	{
		const unsigned mod_ndx( _modified_vertex_indices[ i ] );

		for( unsigned j = 0; j < VertexUsers()[ mod_ndx ].Triangles.size(); ++j )
			_damaged_triangles.push_back( VertexUsers()[ mod_ndx ].Triangles[ j ] );
		for( unsigned j = 0; j < VertexUsers()[ mod_ndx ].Quads.size(); ++j )
			_damaged_quads.push_back( VertexUsers()[ mod_ndx ].Quads[ j ] );
	}
#endif
}

/** Updates modified vertex and polygon normals.
 *
 * Uses the _modified_vertex_indices list to update only modified vertex and
 * polygon normals. This is generally much faster than calling
 * RecalculateAllNormals().
**/
void Mesh::RecalculateModifiedNormals()
{
	unsigned j;
	
	//if( _modified_vertex_indices.size() < VertexLocations().size() )
	if( 1 )
	{
		const DamagedGroup::iterator t_end( _damaged_triangles.end() );
		for( DamagedGroup::iterator a = _damaged_triangles.begin();
		     a != t_end; ++a )
		{
			const unsigned ndx( *a );
			const Triangle& curt( Triangles()[ ndx ] );
			TriangleNormals()[ ndx ] = PlaneUnormal
				( VertexLocations()[ curt.V[ 0 ] ],
				  VertexLocations()[ curt.V[ 1 ] ],
				  VertexLocations()[ curt.V[ 2 ] ] );
		}

		const DamagedGroup::iterator q_end( _damaged_quads.end() );
		for( DamagedGroup::iterator a = _damaged_quads.begin();
		     a != q_end; ++a )
		{
			const unsigned ndx( *a );
			const Quad& curq( Quads()[ ndx ] );
			QuadNormals()[ ndx ] = PlaneUnormal
				( VertexLocations()[ curq.V[ 0 ] ],
				  VertexLocations()[ curq.V[ 1 ] ],
				  VertexLocations()[ curq.V[ 2 ] ] );
		}

		for( unsigned i = 0; i < _modified_vertex_indices.size(); ++i )
		{
			const unsigned mod_ndx( _modified_vertex_indices[ i ] );
			Normal3D& cur_norm( VertexNormals()[ mod_ndx ] );
			const PolygonUsers& usr( VertexUsers()[ mod_ndx ] );

			cur_norm = Origin();
			for( j = 0; j < usr.Triangles.size(); ++j )
				cur_norm += TriangleNormals()[ usr.Triangles[ j ] ];
			for( j = 0; j < usr.Quads.size(); ++j )
				cur_norm += QuadNormals()[ usr.Quads[ j ] ];

			Normalize( cur_norm );
		}
	}
	else
		RecalculateAllNormals();

	/*for( unsigned i = 0; i < _modified_vertex_indices.size(); ++i )
	{
		const unsigned current( _modified_vertex_indices[ i ] );
		const std::set< unsigned > cvs( _connected_vertices( current ) );
		VertexNormals()[ current ].CalculatePlaneNormal
			( VertexLocations()[ *( ++( ++cvs.begin() ) ) ].RawData(),
			  VertexLocations()[ *cvs.begin() ].RawData(),
			  VertexLocations()[ *++cvs.begin() ].RawData() );
		//VertexNormals()[ current ] = Normal3D( 0, 1, 0 );
	}*/

	ChangeOperation( false, false );
}

void Mesh::UpdateVertexUsers()
{
	VertexUsers().clear();
	VertexUsers().resize( VertexLocations().size() );
	for( unsigned i = 0; i < Triangles().size(); ++i )
	{
		const Triangle& t( Triangles()[ i ] );
		VertexUsers()[ t.V[ 0 ] ].Triangles.push_back( i );
		VertexUsers()[ t.V[ 1 ] ].Triangles.push_back( i );
		VertexUsers()[ t.V[ 2 ] ].Triangles.push_back( i );
	}
	for( unsigned i = 0; i < Quads().size(); ++i )
	{
		const Quad& q( Quads()[ i ] );
		VertexUsers()[ q.V[ 0 ] ].Quads.push_back( i );
		VertexUsers()[ q.V[ 1 ] ].Quads.push_back( i );
		VertexUsers()[ q.V[ 2 ] ].Quads.push_back( i );
		VertexUsers()[ q.V[ 3 ] ].Quads.push_back( i );
	}

	update_edges_();
}

void Mesh::update_edges_()
{
	Edges().Edges.clear();
	Edges().Triangles.resize( Triangles().size() );
	Edges().Quads.resize( Quads().size() );

	typedef std::set< unsigned >::const_iterator set_iter;
	typedef std::map< unsigned, unsigned >::const_iterator map_iter;
	for( unsigned i = 0; i < VertexUsers().size(); ++i )
	{
		const std::set< unsigned > con_verts( _connected_vertices( i ) );
		for( set_iter j = con_verts.begin(); j != con_verts.end(); ++j )
		{
			const unsigned cv( *j );
			/* If the 'cv' (connected vertex) has a lesser index than 'i',
			   we know it has already been processed, so ignore. */
			if( cv > i )
			{
				// Find all polygons that use both i and cv
				const std::map< unsigned, unsigned > tri_users( _triangle_users( i, cv ) );
				const std::map< unsigned, unsigned > quad_users( _quad_users( i, cv ) );
				
				for( map_iter k = tri_users.begin(); k != tri_users.end(); ++k )
					Edges().Triangles[ k->first ][ k->second ] = Edges().Edges.size();

				for( map_iter k = quad_users.begin(); k != quad_users.end(); ++k )
					Edges().Quads[ k->first ].E[ k->second ] = Edges().Edges.size();

				Edges().Edges.push_back( Edge( i, cv ) );
			}				
		}
	}
}

std::map< unsigned, unsigned > Mesh::_triangle_users( unsigned v1, unsigned v2 )
{
	std::map< unsigned, unsigned > r;
	// For every user of v1
	for( unsigned i = 0; i < VertexUsers()[ v1 ].Triangles.size(); ++i )
	{
		// For every user of v2
		for( unsigned j = 0; j < VertexUsers()[ v2 ].Triangles.size(); ++j )
		{
			const unsigned poly = VertexUsers()[ v1 ].Triangles[ i ];
			// If poly is a user of both v2 (as well as v1)
			if( poly == VertexUsers()[ v2 ].Triangles[ j ] )
			{
				unsigned w = 0;
				for( unsigned k = 0; k < 3; k++ )
				{
					// Return the side (1, 2, or 3) that's connected
					
					if( Triangles()[ poly ].V[ k ] != v1 &&
					    Triangles()[ poly ].V[ k ] != v2 )
					{
						w = k + 2;
						if( k == 2 )
							w = 1;
						w = k + 1;
						if( w > 2 )
							w -= 3;
						break;
					}
				}
				r[ poly] = w;
				break;
			}
		}
	}
	
	return r;
}
std::map< unsigned, unsigned > Mesh::_quad_users( unsigned v1, unsigned v2 )
{
	std::map< unsigned, unsigned > r;
	// For every user of v1
	for( unsigned i = 0; i < VertexUsers()[ v1 ].Quads.size(); ++i )
	{
		// For every user of v2
		for( unsigned j = 0; j < VertexUsers()[ v2 ].Quads.size(); ++j )
		{
			const unsigned poly = VertexUsers()[ v1 ].Quads[ i ];
			// If poly is a user of both v2 (as well as v1)
			if( poly == VertexUsers()[ v2 ].Quads[ j ] )
			{
				unsigned w = 0;
				for( unsigned k = 0; k < 4; k++ )
				{
					unsigned k2 = k + 1;
					if( k2 > 3 )
						k2 = 0;
					if( ( Quads()[ poly ].V[ k ] == v1 &&
					      Quads()[ poly ].V[ k2 ] == v2 ) ||
					    ( Quads()[ poly ].V[ k ] == v2 &&
					      Quads()[ poly ].V[ k2 ] == v1 ) )
					{
						w = k;
						break;
					}
				}
				r[ poly] = w;
				break;
			}
		}
	}
	
	return r;
}

void Mesh::add_edge_midpoint_( Edge& e )
{
	e.Mid = VertexLocations().size();
	VertexLocations().push_back( Midpoint( VertexLocations()[ e.V[ 0 ] ], VertexLocations()[ e.V[ 1 ] ] ) );
	//std::cout << e.V[ 0 ] << " " << VertexColors()[ e.V[ 0 ] ].R() << " " << VertexColors()[ e.V[ 0 ] ].G() << " " << VertexColors()[ e.V[ 0 ] ].B() << std::endl;
	const Color mix( MixColors( VertexColors()[ e.V[ 0 ] ], VertexColors()[ e.V[ 1 ] ] ) );
	//std::cout << e.V[ 0 ] << " " << mix.R() << " " << mix.G() << " " << mix.B() << std::endl;
	VertexColors().push_back( mix );
	//VertexColors().back().Set( mix.R(), mix.G(), mix.B(), mix.Mask() );
	e.HasMid = true;
}

unsigned Mesh::_poly_count() const
{
	return Triangles().size() + Quads().size();
}

void Mesh::_clear_polys()
{
	Triangles().clear();
	Quads().clear();
	TriangleNormals().clear();
	QuadNormals().clear();
}

void Mesh::ResetVisibles()
{
	_visible_elements.Vertices = VertexLocations().size();
	_visible_elements.Triangles = Triangles().size();
	_visible_elements.Quads = Quads().size();
}

/* Is this function even useful? */
std::set< unsigned > Mesh::_connected_polygon_vertices( unsigned v ) const
{
	std::set< unsigned > r;

	for( unsigned j = 0; j < VertexUsers()[ v ].Triangles.size(); ++j )
	{
		for( unsigned k = 0; k <= 2; ++k )
		{
			const unsigned index = Triangles()[ VertexUsers()[ v ].Triangles[ j ] ].V[ k ];
			if( index != v )
				r.insert( index );
		}
	}
	for( unsigned j = 0; j < VertexUsers()[ v ].Quads.size(); ++j )
	{
		for( unsigned k = 0; k <= 3; ++k )
		{
			const unsigned index = Quads()[ VertexUsers()[ v ].Quads[ j ] ].V[ k ];
			if( index != v )
				r.insert( index );
		}
	}
	return r;
}

/* Find all vertices that are connected by a single edge to the vertex at 'v' */
std::set< unsigned > Mesh::_connected_vertices( unsigned vUS ) const
{
	const int v( vUS );

	std::set< unsigned > r;
	
	for( unsigned j = 0; j < VertexUsers()[ v ].Triangles.size(); ++j )
	{
		for( unsigned k = 0; k <= 2; ++k )
		{
			if( v == static_cast< signed >( Triangles()[ VertexUsers()[ v ].Triangles[ j ] ].V[ k ] ) )
			{
				const unsigned add( k + 1 );

				r.insert( Triangles()[ VertexUsers()[ v ].Triangles[ j ] ].V[ k == 0 ? 2 : k - 1 ] );
				r.insert( Triangles()[ VertexUsers()[ v ].Triangles[ j ] ].V[ add > 2 ? 0 : add ] );

				break;
			}
		}
	}
	for( unsigned j = 0; j < VertexUsers()[ v ].Quads.size(); ++j )
	{
		for( unsigned k = 0; k <= 3; ++k )
		{
			if( v == static_cast< signed >( Quads()[ VertexUsers()[ v ].Quads[ j ] ].V[ k ] ) )
			{
				const unsigned add( k + 1 );

				/*std::cout << "qsize=" << Quads().size() << " qi=" << VertexUsers()[ v ].Quads[ j ]
					  << " k=" << k
					  << " v1=" << ( k == 0 ? 3 : k - 1 )
					  << " v2=" << ( add > 3 ? 0 : add ) << std::endl;*/

				r.insert( Quads()[ VertexUsers()[ v ].Quads[ j ] ].V[ k == 0 ? 3 : k - 1 ] );
				r.insert( Quads()[ VertexUsers()[ v ].Quads[ j ] ].V[ add > 3 ? 0 : add ] );

				break;
			}
		}
	}
	
	return r;
}

/** Stores vertices within the MainBrush->GetRadius().
 *
 * Stores a list of all vertices within the sphere that has its center at area
 * and a radius of MainBrush->GetRadius(). The distance from the active vertices
 * to the center is also stored.
**/
// Slow
/*void Mesh::_find_active_vertices( std::vector< unsigned > &vertex_indices,
                                  std::vector< float > &vertex_distances,
                                  EditData e )
{
	vertex_indices.clear();
	vertex_distances.clear();
	
	float distance;
	
	for( int i = 0; i < _visible_elements.Vertices; ++i )
	{
		distance = e.Center.Distance( VertexLocations()[ i ] );
		if( distance < _selection_radius )
		{
			vertex_indices.push_back( i );
			vertex_distances.push_back( distance );
		}
	}
	
	_selected_vertex[ e.SymmetryIndex ] = vertex_indices[ 0 ];

	_modified_vertex_indices.insert( _modified_vertex_indices.end(),
					 vertex_indices.begin(),
	                                 vertex_indices.end() );
}*/
/* This version of _find_active_vertices may be rather slow too...*/
/*void Mesh::_find_active_vertices( std::vector< unsigned > &vertex_indices,
                                  std::vector< float > &vertex_distances,
                                  EditData e )
{	
	vertex_indices.clear();
	vertex_distances.clear();
	
	int area_index = 0;

	if( _selected_vertex[ e.SymmetryIndex ] >= VertexLocations().size() )
		_selected_vertex[ e.SymmetryIndex ] = 0;

	// If a point used last time through the loop is within the Radius, use it!
	if( VertexLocations()[ _selected_vertex[ e.SymmetryIndex ] ].Distance( e.Center ) <
	    _selection_radius )
	{
		area_index = _selected_vertex[ e.SymmetryIndex ];
	}
	else
	{
		float tmp_distance;
		
		// The point of this loop is just to get any vertex that is within the Radius
		for( int i = 0; i < VertexLocations().size(); ++i )
		{
			e.Center.FastDistance( VertexLocations()[ i ], tmp_distance );

			if( tmp_distance < _selection_radius )
			{
				area_index = i;
				break;
			}
		}
	}

	std::set< unsigned > checked;
	ActiveVertexData d( area_index, checked, vertex_indices, vertex_distances, e.Center );
	_is_vertex_active( d );
	
	_selected_vertex[ e.SymmetryIndex ] = vertex_indices[ 0 ];

	_modified_vertex_indices.insert( _modified_vertex_indices.end(),
					 vertex_indices.begin(),
	                                 vertex_indices.end() );
}*/

/*void Mesh::_is_vertex_active( ActiveVertexData& d )
{
	unsigned current = d.Current;

	int xdf = d.Checked.size();
	d.Checked.insert( current );
	
	float distance = d.Area.Distance( VertexLocations()[ current ] );
	if( distance < _selection_radius )
	{
		if( current < _lowest_index )
			_lowest_index = current;
		if( current > _highest_index )
			_highest_index = current;
		d.VertexIndices.push_back( current );
		d.VertexDistances.push_back( distance );
	}
	else
		return;
	
	int j, polygon_index;
	int previous_point_index;
	// Loop through each triangle using the current point
	for( unsigned i = 0; i < VertexUsers()[ current ].Triangles.size(); ++i )
	{
		// Set the current polygon
		polygon_index =VertexUsers()[ current ].Triangles[ i ];
		// Loop through each vertex in the polygon
		for( j = 0; j <= 2; ++j )
			// If the current active point equals the current VertexIndex...
			if( Triangles()[ polygon_index ].V[ j ] == current )
			{
				previous_point_index = j - 1;
				if( previous_point_index < 0 )
					previous_point_index = 2;

				previous_point_index = Triangles()[ polygon_index ].V[ previous_point_index ];

				if( d.Checked.find( previous_point_index ) == d.Checked.end() )
				{
					d.Current = previous_point_index;
					_is_vertex_active( d );
				}
				break;
			}
	}
	
	// Loop through each quad using the current point
	for( unsigned i = 0; i < VertexUsers()[ current ].Quads.size(); ++i )
	{
		// Set the current polygon
		polygon_index =VertexUsers()[ current ].Quads[ i ];
		// Loop through each vertex in the polygon
		for( j = 0; j <= 3; ++j )
			// If the current active point equals the current VertexIndex...
			if( Quads()[ polygon_index ].V[ j ] == current )
			{
				previous_point_index = j - 1;
				if( previous_point_index < 0 )
					previous_point_index = 3;

				previous_point_index = Quads()[ polygon_index ].V[ previous_point_index ];

				if( d.Checked.find( previous_point_index ) == d.Checked.end() )
				{
					d.Current = previous_point_index;
					_is_vertex_active( d );
				}
				break;
			}
	}
}
*/
