/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Mesh.hh"
#include "MeshStepables.hh"

using SharpConstruct::FinalizeCreation;
using SharpConstruct::UpdateNormals;
using SharpConstruct::Mesh;

FinalizeCreation::FinalizeCreation( Mesh& mesh )
: mesh_( mesh ), update_normals_( mesh ),
  step_one_done_( false ), step_two_done_( false ), done_( false )
{}

double FinalizeCreation::Progress() const
{
	if( !step_one_done_ )
		return 0;
	else if( !step_two_done_ )
		//return update_normals_.Progress();
		return 1;
	else
		return 1;
}

const std::string FinalizeCreation::Text() const
{
	if( !step_one_done_ )
		return "Saning...";
	if( !step_two_done_ )
		return "Updating normals...";

	return "Finalizing...";
}

bool FinalizeCreation::Step()
{
	if( done_ )
		return DONE;

	if( !step_one_done_ )
	{
		const unsigned vertices( mesh_.VertexLocations().size() );
		mesh_.VertexNormals().resize( vertices );
		mesh_.VertexColors().resize( vertices );
		mesh_.TriangleNormals().resize( mesh_.Triangles().size() );
		mesh_.QuadNormals().resize( mesh_.Quads().size() );
		mesh_.ResetVisibles();
		mesh_.UpdateVertexUsers();
		//mesh__saved = true;

		step_one_done_ = true;
		return CONTINUE;
	}
	else if( !step_two_done_ )
	{
		//mesh_.Unify();
		/*if( modifiable )
		{
			Unify();
			Flood( Color( 1, 1, 1 ), 1 );
		}*/

		/*if( update_normals_.Step() )
			return CONTINUE;
			else*/
		mesh_.RecalculateAllNormals();
			step_two_done_ = true;
		return CONTINUE;
	}
	else
	{
		mesh_.UpdateBounds();
		//Mesh::CurrentGrid().SetDivision( 10 );
		mesh_.ChangeOperation( true, true );
		done_ = true;
		return DONE;
	}
}

UpdateNormals::UpdateNormals( Mesh& mesh )
: mesh_( mesh ), current_triangle_( 0 ), current_quad_( 0 ),
  current_vertex_( 0 )
{}

double UpdateNormals::Progress() const
{
	const unsigned den( mesh_.Triangles().size() + mesh_.Quads().size() + mesh_.VertexLocations().size() );
	if( den )
		return ( current_triangle_ + current_quad_ + current_vertex_ ) * 1.0 / den;
	else
		return 0;
}

const std::string UpdateNormals::Text() const
{
	return "Updating normals...";
}

bool UpdateNormals::Step()
{
	if( !step_triangles_() )
		if( !step_quads_() )
			return step_vertices_();
	
	return CONTINUE;
}

bool UpdateNormals::step_triangles_()
{
	Point3DVector& normals( mesh_.TriangleNormals() );
	if( current_triangle_ < normals.size() )
	{
		const Point3DVector& vertices( mesh_.VertexLocations() );
		const Triangle& poly( mesh_.Triangles()[ current_triangle_ ] );
		normals[ current_triangle_ ] = PlaneUnormal
			( vertices[ poly.V[ 0 ] ],
			  vertices[ poly.V[ 1 ] ],
			  vertices[ poly.V[ 2 ] ] );
		++current_triangle_;
		return CONTINUE;
	}
	else
		return DONE;
}

bool UpdateNormals::step_quads_()
{
	Point3DVector& normals( mesh_.QuadNormals() );
	if( current_quad_ < normals.size() )
	{
		const Point3DVector& vertices( mesh_.VertexLocations() );
		const Quad& poly( mesh_.Quads()[ current_quad_ ] );
		normals[ current_quad_ ] = PlaneUnormal
			( vertices[ poly.V[ 0 ] ],
			  vertices[ poly.V[ 1 ] ],
			  vertices[ poly.V[ 2 ] ] );
		++current_quad_;
		return CONTINUE;
	}
	else
		return DONE;
}

bool UpdateNormals::step_vertices_()
{
	const std::vector< PolygonUsers >& users( mesh_.VertexUsers() );
	if( current_vertex_ < users.size() )
	{
		Normal3D& cur_norm( mesh_.VertexNormals()[ current_vertex_ ] );
		const PolygonUsers& usr( users[ current_vertex_ ] );
		cur_norm = Origin();
		for( unsigned j = 0; j < usr.Triangles.size(); ++j )
			cur_norm += mesh_.TriangleNormals()[ usr.Triangles[ j ] ];
		for( unsigned k = 0; k < usr.Quads.size(); k++ )
			cur_norm += mesh_.QuadNormals()[ usr.Quads[ k ] ];

		Normalize( cur_norm );

		++current_vertex_;
		return CONTINUE;
	}
	else
		return DONE;
}
