/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "MeshHistory.hh"
#include "Move.hh"

#ifdef WIN32
#include <windows.h>
#include "win32.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>

using SharpConstruct::Move;

Move::Move()
{
	transforms_[ X ] = transforms_[ Y ] = transforms_[ Z ] =
		transforms_[ ALL ] = 0;
	final_transforms_ = transforms_;
}

void Move::ModifyModel()
{
	const float sf( 1 );
	glTranslatef( transforms_[ X ] +
		      final_transforms_[ X ] * sf,
		      transforms_[ Y ] +
		      final_transforms_[ Y ] * sf,
		      transforms_[ Z ] +
		      final_transforms_[ Z ] * sf );
}

void Move::MouseMove( const Gdk::Point& mloc )
{
	float dist( 0 );
	if( selected_ != NONE )
	{
		Math::Point2D< float > closest;
		{
			const Math::Line2D< int > line( model_center_, handle_down_locs_[ selected_ ] );

			float slope1, slope2;
						
			if( abs( line.Start().X() - line.End().X() ) < 0.01 )
			{
				closest.SetX( line.Start().X() );
				closest.SetY( mloc.get_y() );
			}
			else if( abs( line.Start().Y() - line.End().Y() ) < 0.01 )
			{
				closest.SetX( mloc.get_x() );
				closest.SetY( line.Start().Y() );
			}
			else
			{
				slope1 = float( line.Start().Y() - line.End().Y() ) / ( line.Start().X() - line.End().X() );
				slope2 = -1 / slope1;
				closest.SetX( ( -slope2 * mloc.get_x() + mloc.get_y() - line.Start().Y() + slope1 * line.Start().X()) /
					      ( slope1 - slope2 ) );
				closest.SetY( slope1 * ( closest.X() - line.Start().X() ) + line.Start().Y() );
			}
		}

		dist = Math::Distance( Gdk::Point( lroundf( closest.X() ), lroundf( closest.Y() ) ),
				       model_center_ );

		// See if dist should be negative
		{
			const float norm( -1.0f * ( model_center_.get_x() - handle_down_locs_[ selected_ ].get_x() ) /
					  ( model_center_.get_y() - handle_down_locs_[ selected_ ].get_y() ) );
			const float b( model_center_.get_y() - norm * model_center_.get_x() );
			
			if( mloc.get_y() > norm * mloc.get_x() + b )
				dist *= -1;
			if( model_center_.get_y() < handle_down_locs_[ selected_ ].get_y() )
				dist *= -1;
			if( ( model_center_.get_y() == handle_down_locs_[ selected_ ].get_y() ) &&
			    ( mloc.get_x() < model_center_.get_x() ) )
				dist *= -1;
		}
	}
	
	transforms_[ selected_ ] = dist / Math::Distance( model_center_, mouse_down_loc_ ) - 1;
}

void Move::MouseUp( const Gdk::Point& )
{
	for( std::map< Axis, float >::iterator i = transforms_.begin();
		     i != transforms_.end(); ++i )
	{
		final_transforms_[ i->first ] += i->second;
		i->second = 0;
	}

	selected_ = NONE;
}

void Move::Apply()
{
	MeshHistory::Instance().AddMesh( true, false, false );
	MeshHistory::Instance().GetCurrentMesh().Move( transforms_[ X ] +
						       final_transforms_[ X ],
						       transforms_[ Y ] +
						       final_transforms_[ Y ],
						       transforms_[ Z ] +
						       final_transforms_[ Z ] );
}

void Move::draw_handle_( const Axis a, const float x, const float y, const float z )
{
	const float delta( 10 );

	GLUquadric* quadric = gluNewQuadric();

	glPushMatrix();
	glRotatef( 90, x, -y, z );
	glBegin( GL_LINES );
	glVertex3f( 0, 0, 0 );
	glVertex3f( delta, 0, 0 );
	glEnd();
	glTranslatef( delta, 0, 0 );
	glRotatef( 90, 0, 1, 0 );
	gluCylinder( quadric, 1, 0, 2, 10, 1 );
	handle_down_locs_[ a ] = project_2d_( Origin() );
	glPopMatrix();

	gluDeleteQuadric( quadric );
}
