/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "MeshHistory.hh"
#include "Scale.hh"

#ifdef WIN32
#include <windows.h>
#include "win32.h"
#endif

#include <GL/gl.h>

using SharpConstruct::Scale;

Scale::Scale()
{
	transforms_[ X ] = transforms_[ Y ] = transforms_[ Z ] =
		transforms_[ ALL ] = 1;
	final_transforms_ = transforms_;
}

void Scale::ModifyModel()
{
	glScalef( final_transforms_[ X ] * final_transforms_[ ALL ] *
		  transforms_[ X ] * transforms_[ ALL ],
		  final_transforms_[ Y ] * final_transforms_[ ALL ] *
		  transforms_[ Y ] * transforms_[ ALL ],
		  final_transforms_[ Z ] * final_transforms_[ ALL ] *
		  transforms_[ Z ] * transforms_[ ALL ] );
}

void Scale::MouseMove( const Gdk::Point& mloc )
{
	float dist( 0 );
	
	if( selected_ == ALL )
		dist = Math::Distance( mloc, model_center_ );
	else if( selected_ != NONE )
	{
		Math::Point2D< float > closest;
		{
			const Math::Line2D< int > line( model_center_, handle_down_locs_[ selected_ ] );

			float slope1, slope2;
						
			if( abs( line.Start().X() - line.End().X() ) < 0.01 )
			{
				closest.SetX( line.Start().X() );
				closest.SetY( mloc.get_y() );
			}
			else if( abs( line.Start().Y() - line.End().Y() ) < 0.01 )
			{
				closest.SetX( mloc.get_x() );
				closest.SetY( line.Start().Y() );
			}
			else
			{
				slope1 = float( line.Start().Y() - line.End().Y() ) / ( line.Start().X() - line.End().X() );
				slope2 = -1 / slope1;
				closest.SetX( ( -slope2 * mloc.get_x() + mloc.get_y() - line.Start().Y() + slope1 * line.Start().X()) /
					      ( slope1 - slope2 ) );
				closest.SetY( slope1 * ( closest.X() - line.Start().X() ) + line.Start().Y() );
			}
		}

		dist = Math::Distance( Gdk::Point( lroundf( closest.X() ), lroundf( closest.Y() ) ),
				       model_center_ );
	}

	if( selected_ != NONE )
		transforms_[ selected_ ] = dist / Math::Distance( model_center_, mouse_down_loc_ );
}

void Scale::MouseUp( const Gdk::Point& )
{
	for( std::map< Axis, float >::iterator i = transforms_.begin();
		     i != transforms_.end(); ++i )
	{
		final_transforms_[ i->first ] *= i->second;
		i->second = 1;
	}

	selected_ = NONE;
}

void Scale::Apply()
{
	MeshHistory::Instance().AddMesh( true, false, false );
	MeshHistory::Instance().GetCurrentMesh().Resize( final_transforms_[ X ] * final_transforms_[ ALL ],
							 final_transforms_[ Y ] * final_transforms_[ ALL ],
							 final_transforms_[ Z ] * final_transforms_[ ALL ] );
}

void Scale::draw_handle_( const Axis a, const float x, const float y, const float z )
{
	const float delta( 10 );

	glPushMatrix();
	glRotatef( 90, x, -y, z );
	glBegin( GL_LINES );
	glVertex3f( 0, 0, 0 );
	glVertex3f( delta, 0, 0 );
	glEnd();
	glTranslatef( delta, 0, 0 );
	draw_cube_();
	handle_down_locs_[ a ] = project_2d_( Origin() );
	glPopMatrix();
}

void Scale::draw_cube_() const
{
	glBegin( GL_QUADS );
	{
		const float l( 1 );
		
		// Front Face
		//glNormal3f( 0,  0,  1 );
		glVertex3f(-l, -l,  l );
		glVertex3f( l, -l,  l );
		glVertex3f( l,  l,  l );
		glVertex3f(-l,  l,  l );
		
		// Back Face
		//glNormal3f( 0,  0, -1 );
		glVertex3f(-l, -l, -l );
		glVertex3f(-l,  l, -l );
		glVertex3f( l,  l, -l );
		glVertex3f( l, -l, -l );
		
		// Top Face
		//glNormal3f( 0,  1,  0 );
		glVertex3f(-l,  l, -l );
		glVertex3f(-l,  l,  l );
		glVertex3f( l,  l,  l );
		glVertex3f( l,  l, -l );
		
		// Bottom Face
		//glNormal3f( 0, -1,  0 );
		glVertex3f(-l, -l, -l );
		glVertex3f( l, -l, -l );
		glVertex3f( l, -l,  l );
		glVertex3f(-l, -l,  l );
		
		// Right face
		//glNormal3f( 1,  0,  0 );
		glVertex3f( l, -l, -l );
		glVertex3f( l,  l, -l );
		glVertex3f( l,  l,  l );
		glVertex3f( l, -l,  l );
		
		// Left Face
		//glNormal3f(-1,  0,  0 );
		glVertex3f(-l, -l, -l );
		glVertex3f(-l, -l,  l );
		glVertex3f(-l,  l,  l );
		glVertex3f(-l,  l, -l );
	}
	glEnd();
}
